#ifndef COGALU_H
#define COGALU_H

#include <QString>

enum PropellerInstruction {
    InstructionHubByte    = 0x00,
    InstructionHubWord    = 0x01,
    InstructionHubLong    = 0x02,
    InstructionSysOp      = 0x03,
    InstructionIllgMul    = 0x04,
    InstructionIllgMuls   = 0x05,
    InstructionIllgEnc    = 0x06,
    InstructionIllgOnes   = 0x07,

    InstructionRor        = 0x08,
    InstructionRol        = 0x09,
    InstructionShr        = 0x0A,
    InstructionShl        = 0x0B,
    InstructionRcr        = 0x0C,
    InstructionRcl        = 0x0D,
    InstructionSar        = 0x0E,
    InstructionRev        = 0x0F,

    InstructionMinS       = 0x10,
    InstructionMaxS       = 0x11,
    InstructionMin        = 0x12,
    InstructionMax        = 0x13,
    InstructionMovS       = 0x14,
    InstructionMovD       = 0x15,
    InstructionMovI       = 0x16,
    InstructionJmpRet     = 0x17,

    InstructionAnd        = 0x18,
    InstructionAndN       = 0x19,
    InstructionOr         = 0x1A,
    InstructionXor        = 0x1B,
    InstructionMuxC       = 0x1C,
    InstructionMuxNC      = 0x1D,
    InstructionMuxZ       = 0x1E,
    InstructionMuxNZ      = 0x1F,

    InstructionAdd        = 0x20,
    InstructionSub        = 0x21,
    InstructionAddAbs     = 0x22,
    InstructionSubAbs     = 0x23,
    InstructionSumC       = 0x24,
    InstructionSumNC      = 0x25,
    InstructionSumZ       = 0x26,
    InstructionSumNZ      = 0x27,

    InstructionMov        = 0x28,
    InstructionNeg        = 0x29,
    InstructionAbs        = 0x2A,
    InstructionAbsNeg     = 0x2B,
    InstructionNegC       = 0x2C,
    InstructionNegNC      = 0x2D,
    InstructionNegZ       = 0x2E,
    InstructionNegNZ      = 0x2F,

    InstructionCmpS       = 0x30,
    InstructionCmpSX      = 0x31,
    InstructionAddX       = 0x32,
    InstructionSubX       = 0x33,
    InstructionAddS       = 0x34,
    InstructionSubS       = 0x35,
    InstructionAddSX      = 0x36,
    InstructionSubSX      = 0x37,

    InstructionCmpSub     = 0x38,
    InstructionDJNZ       = 0x39,
    InstructionTJNZ       = 0x3A,
    InstructionTJZ        = 0x3B,
    InstructionWaitPEq    = 0x3C,
    InstructionWaitPNe    = 0x3D,
    InstructionWaitCnt    = 0x3E,
    InstructionWaitVid    = 0x3F,
};

struct CogAluResult {
    quint32 result;
    bool writeEnable;
    bool carryOut;
    bool zeroOut;

    CogAluResult(PropellerInstruction instrCode, quint32 sVal, quint32 dVal, quint32 instrPtr, bool run, bool carryIn, bool zeroIn, quint32 bus_q, bool bus_c):
        result(0),
        writeEnable(true),
        carryOut(false),
        zeroOut(false)
    {
        switch(instrCode) {
            case InstructionHubByte:
            case InstructionHubWord:
            case InstructionHubLong:
            case InstructionSysOp:
            case InstructionIllgMul:
            case InstructionIllgMuls:
            case InstructionIllgEnc:
            case InstructionIllgOnes:
                result = (run || ((instrPtr&0x1F0) != 0x1F0)) ? bus_q : 0;
                carryOut = bus_c;
                zeroOut = result==0;
                break;
            case InstructionRor:
                result = shiftRight(dVal,sVal) | shiftLeftInv(dVal,sVal);
                zeroOut = result==0;
                carryOut = (dVal & 1);
                break;
            case InstructionRol:
                result = shiftLeft(dVal,sVal) | shiftRightInv(dVal,sVal);
                zeroOut = result==0;
                carryOut = (dVal & 0x80000000) != 0;
                break;
            case InstructionShr:
                result = shiftRight(dVal,sVal);
                zeroOut = result==0;
                carryOut = (dVal & 1);
                break;
            case InstructionShl:
                result = shiftLeft(dVal,sVal);
                zeroOut = result==0;
                carryOut = (dVal & 0x80000000) != 0;
                break;
            case InstructionRcr:
                result = shiftRight(dVal,sVal);
                if (carryIn)
                    result |= shiftLeftInv(0xFFFFFFFF,sVal);
                zeroOut = result==0;
                carryOut = (dVal & 1);
                break;
            case InstructionRcl:
                result = shiftLeft(dVal,sVal);
                if (carryIn)
                    result |= shiftRightInv(0xFFFFFFFF,sVal);
                zeroOut = result==0;
                carryOut = (dVal & 0x80000000) != 0;
                break;
            case InstructionSar:
                result = shiftRight(dVal,sVal);
                if (valueIsNegative(dVal))
                    result |= shiftLeftInv(0xFFFFFFFF,sVal);
                zeroOut = result==0;
                carryOut = (dVal & 1);
                break;
            case InstructionRev:
                result = shiftRight(reverseBits(dVal),sVal);
                zeroOut = result==0;
                carryOut = (dVal & 1);
                break;

            case InstructionMinS:
                carryOut = signedIsLess(dVal,sVal);
                result = sVal;
                writeEnable = carryOut;
                zeroOut = result==0;
                break;
            case InstructionMaxS:
                carryOut = signedIsLess(dVal,sVal);
                result = sVal;
                writeEnable = !carryOut;
                zeroOut = result==0;
                break;
            case InstructionMin:
                carryOut = dVal<sVal;
                result = sVal;
                writeEnable = carryOut;
                zeroOut = result==0;
                break;
            case InstructionMax:
                carryOut = dVal<sVal;
                result = sVal;
                writeEnable = !carryOut;
                zeroOut = result==0;
                break;
            case InstructionMovS:
                carryOut = dVal<sVal; //undokumentierter seiteneffekt
                result = (dVal & 0xFFFFFE00) | (sVal & 0x1FF);
                zeroOut = result==0;
                break;
            case InstructionMovD:
                carryOut = dVal<sVal; //undokumentierter seiteneffekt
                result = (dVal & 0xFFFC01FF) | ((sVal & 0x1FF) << 9);
                zeroOut = result==0;
                break;
            case InstructionMovI:
                carryOut = dVal<sVal; //undokumentierter seiteneffekt
                result = (dVal & 0x007FFFFF) | ((sVal & 0x1FF) << 23);
                zeroOut = result==0;
                break;
            case InstructionJmpRet:
                carryOut = dVal<sVal; //undokumentierter seiteneffekt
                result = (dVal & 0xFFFFFE00) | instrPtr;
                zeroOut = result==0;
                break;
            case InstructionAnd:
                result = dVal & sVal;
                carryOut = getParity(result);
                zeroOut = result==0;
                break;
            case InstructionAndN:
                result = dVal & ~sVal;
                carryOut = getParity(result);
                zeroOut = result==0;
                break;
            case InstructionOr:
                result = dVal | sVal;
                carryOut = getParity(result);
                zeroOut = result==0;
                break;
            case InstructionXor:
                result = dVal ^ sVal;
                carryOut = getParity(result);
                zeroOut = result==0;
                break;
            case InstructionMuxC:
                result = carryIn ? (dVal | sVal) : (dVal & ~sVal);
                carryOut = getParity(result);
                zeroOut = result==0;
                break;
            case InstructionMuxNC:
                result = !carryIn ? (dVal | sVal) : (dVal & ~sVal);
                carryOut = getParity(result);
                zeroOut = result==0;
                break;
            case InstructionMuxZ:
                result = zeroIn ? (dVal | sVal) : (dVal & ~sVal);
                carryOut = getParity(result);
                zeroOut = result==0;
                break;
            case InstructionMuxNZ:
                result = !zeroIn ? (dVal | sVal) : (dVal & ~sVal);
                carryOut = getParity(result);
                zeroOut = result==0;
                break;
            case InstructionAdd:
                unsignedAddOrSub(false,false,sVal,dVal);
                zeroOut = result==0;
                break;
            case InstructionSub:
                unsignedAddOrSub(true,false,sVal,dVal);
                zeroOut = result==0;
                break;
            case InstructionAddAbs: {
                const quint64 tmpResult = valueIsNegative(sVal) ? extendToU64(dVal)+extendToU64(-sVal) : extendToU64(dVal)+extendToU64(sVal);
                carryOut = valueIsNegative(sVal) ? !getBit32(tmpResult) : getBit32(tmpResult);
                result = getLower32FromU64(tmpResult);
                zeroOut = result==0;
                break;
            }
            case InstructionSubAbs: {
                const quint64 tmpResult = valueIsNegative(sVal) ? extendToU64(dVal)-extendToU64(-sVal) : extendToU64(dVal)-extendToU64(sVal);
                carryOut = valueIsNegative(sVal) ? !getBit32(tmpResult) : getBit32(tmpResult);
                result = getLower32FromU64(tmpResult);
                zeroOut = result==0;
                break;
            }
            case InstructionSumC:
                signedAddOrSub(carryIn,false,sVal,dVal);
                zeroOut = result==0;
                break;
            case InstructionSumNC:
                signedAddOrSub(!carryIn,false,sVal,dVal);
                zeroOut = result==0;
                break;
            case InstructionSumZ:
                signedAddOrSub(zeroIn,false,sVal,dVal);
                zeroOut = result==0;
                break;
            case InstructionSumNZ:
                signedAddOrSub(!zeroIn,false,sVal,dVal);
                zeroOut = result==0;
                break;
            case InstructionMov:
                conditionallyNegate(0,sVal);
                break;
            case InstructionNeg:
                conditionallyNegate(1,sVal);
                break;
            case InstructionAbs:
                conditionallyNegate(valueIsNegative(sVal),sVal);
                break;
            case InstructionAbsNeg:
                conditionallyNegate(!valueIsNegative(sVal),sVal);
                break;
            case InstructionNegC:
                conditionallyNegate(carryIn,sVal);
                break;
            case InstructionNegNC:
                conditionallyNegate(!carryIn,sVal);
                break;
            case InstructionNegZ:
                conditionallyNegate(zeroIn,sVal);
                break;
            case InstructionNegNZ:
                conditionallyNegate(!zeroIn,sVal);
                break;
            case InstructionCmpS:
                signedAddOrSub(true,false,sVal,dVal); //nötig damit das result richtig ist (undokumentiert)
                carryOut = ((result>>31)&1) ^ carryOut;
                zeroOut = result==0;
                break;
            case InstructionCmpSX:
                signedAddOrSub(true,carryIn,sVal,dVal); //nötig damit das result richtig ist (undokumentiert)
                carryOut = ((result>>31)&1) ^ carryOut;
                zeroOut = zeroIn && (result==0);
                break;
            case InstructionAddX:
                unsignedAddOrSub(false,carryIn,sVal,dVal);
                zeroOut = zeroIn && (result==0);
                break;
            case InstructionSubX:
                unsignedAddOrSub(true,carryIn,sVal,dVal);
                zeroOut = zeroIn && (result==0);
                break;
            case InstructionAddS:
                signedAddOrSub(false,false,sVal,dVal);
                zeroOut = result==0;
                break;
            case InstructionSubS:
                signedAddOrSub(true,false,sVal,dVal);
                zeroOut = result==0;
                break;
            case InstructionAddSX:
                signedAddOrSub(false,carryIn,sVal,dVal);
                zeroOut = (result==0) & zeroIn;
                break;
            case InstructionSubSX:
                signedAddOrSub(true,carryIn,sVal,dVal);
                zeroOut = (result==0) & zeroIn;
                break;
            case InstructionCmpSub:
                unsignedAddOrSub(true,false,sVal,dVal);
                carryOut =sVal<=dVal;
                writeEnable = carryOut;
                zeroOut = result==0;
                break;
            case InstructionDJNZ:
                result = dVal-1;
                zeroOut = result==0;
                carryOut = dVal==0;
                break;
            case InstructionTJNZ:
                result = dVal;
                zeroOut = result==0;
                carryOut = 0;
                break;
            case InstructionTJZ:
                result = dVal;
                zeroOut = result==0;
                carryOut = 0;
                break;
            case InstructionWaitPEq:
                unsignedAddOrSub(false,false,sVal,dVal);
                zeroOut = (result==0);
                break;
            case InstructionWaitPNe:
                unsignedAddOrSub(false,1,sVal,dVal);
                zeroOut = (result==0);
                break;
            case InstructionWaitCnt:
                unsignedAddOrSub(false,false,sVal,dVal);
                zeroOut = (result==0);
                break;
            case InstructionWaitVid:
                unsignedAddOrSub(false,false,sVal,dVal);
                zeroOut = (result==0);
                break;
        }
    }

    static bool valueIsNegative(quint32 value) {
        return (value & 0x80000000) != 0x0;
    }

    static bool getParity(quint32 value) {
        value ^= (value >> 16);
        value ^= (value >> 8);
        value ^= (value >> 4);
        value ^= (value >> 2);
        return ((value ^ (value >> 1)) & 1)==1;
    }

    static quint32 reverseBits(quint32 value) {
        value = (((value & 0xaaaaaaaa) >> 1) | ((value & 0x55555555) << 1));
        value = (((value & 0xcccccccc) >> 2) | ((value & 0x33333333) << 2));
        value = (((value & 0xf0f0f0f0) >> 4) | ((value & 0x0f0f0f0f) << 4));
        value = (((value & 0xff00ff00) >> 8) | ((value & 0x00ff00ff) << 8));
        return ((value >> 16) | (value << 16));
    }

    static quint64 extendToU64(quint32 value) {
        return static_cast<quint64>(value);
    }

    static quint64 extendToU64(bool value) {
        return value ? 1L : 0L;
    }

    static quint64 extendToU64ShiftSignToUpper(quint32 value) {
        return ((extendToU64(value) & 0x80000000L) << 1) | (extendToU64(value) & 0x7FFFFFFFL);
    }

    static quint32 getLower32FromU64(quint64 value) {
        return static_cast<quint32>(value & 0xFFFFFFFF);
    }

    static bool getBit32(quint64 value) {
        return ((value >> 32) & 1) == 1;
    }

    static bool signedIsLess(quint32 left, quint32 right) {
        return qint32(left) < qint32(right);
    }

    static quint32 shiftLeft(quint32 value, quint32 shiftBy) {
        shiftBy &= 31;
        return value << shiftBy;
    }

    static quint32 shiftLeftInv(quint32 value, quint32 shiftBy) {
        shiftBy = 32-(shiftBy & 31);
        return shiftBy == 32 ? 0 : value << shiftBy;
    }

    static quint32 shiftRight(quint32 value, quint32 shiftBy) {
        shiftBy &= 31;
        return value >> shiftBy;
    }

    static quint32 shiftRightInv(quint32 value, quint32 shiftBy) {
        shiftBy = 32-(shiftBy & 31);
        return shiftBy == 32 ? 0 : value >> shiftBy;
    }

    void conditionallyNegate(bool doInversion, quint32 value) {
        result = doInversion ? -value : value;
        zeroOut = result==0;
        carryOut = valueIsNegative(value);
    }

    void unsignedAddOrSub(bool doSubtraction, bool addExtraCarry, quint32 sVal, quint32 dVal) {
        const quint64 tmpResult = extendToU64(dVal)+
        (doSubtraction
            ? -(extendToU64(sVal)+extendToU64(addExtraCarry))
            :  (extendToU64(sVal)+extendToU64(addExtraCarry)));
        result = getLower32FromU64(tmpResult);
        carryOut = getBit32(tmpResult);
    }

    void signedAddOrSub(bool doSubtraction, bool addExtraCarry, quint32 sVal, quint32 dVal) {
        const quint64 tmpResult = extendToU64ShiftSignToUpper(dVal)
        + extendToU64ShiftSignToUpper(doSubtraction ? ~sVal : sVal)
        + (doSubtraction==addExtraCarry ? 0x80000000L : 0x80000001L);

        result = ((tmpResult >>1)&0x80000000L) | (tmpResult & 0x7FFFFFFFL);
        carryOut = ((tmpResult >> 33)&1) == ((tmpResult >> 31)&1);
    }
};

#endif // COGALU_H
