#include "cogviewer.h"
#include "ui_cogviewer.h"
#include "pdisasm.h"
#include "cogramannotationdialog.h"
#include <QtDebug>
#include <QPair>
#include <QFileDialog>

CogViewer::CogViewer(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::CogViewer),
    m_propeller(nullptr),
    m_cogNo(0)
{
    ui->setupUi(this);

    m_annotationFileName = "other-files/annotation-loader.txt";
    loadAnnotation();
}

CogViewer::~CogViewer()
{
    delete ui;
}

void CogViewer::setPropeller(AbstractPropeller *propeller, int cogNo) {
    m_propeller = propeller;
    m_cogNo = cogNo;
    updateState(true);
}

QVector<bool> CogViewer::getStepOverList() const {
    QVector<bool> result(m_annotation.annotations.size());
    for (int i=0; i<result.size(); ++i)
        result[i] = m_annotation.annotations[i].stepOver;
    return result;
}

void CogViewer::updateState(bool resetCurrentIndex) {
    if (!m_propeller) {
        ui->lblStatus->clear();
        ui->lstCogMemory->clear();
        ui->lstCogInnerVars->clear();
        return;
    }
    auto status = m_propeller->cogStatus(m_cogNo);
    //Update Cog Memory View
    QList<QPair<int,QString>> newItems;
    for (int i=0; i<512; ++i) {
        for (int j=0; j<m_annotation.annotations[i].prependComments.size(); ++j)
            newItems << QPair<int,QString>(i, "# "+m_annotation.annotations[i].prependComments[j]);
        newItems << QPair<int,QString>(i, getTextForCogRamEntry(i,m_propeller->cogMemory(m_cogNo, i),i == status.programCounter));
    }
    if (newItems.size()<ui->lstCogMemory->count())
        ui->lstCogMemory->clear();
    for (int i=0; i<newItems.size(); ++i) {
        if (i>=ui->lstCogMemory->count())
            ui->lstCogMemory->addItem(QString());
        ui->lstCogMemory->item(i)->setData(Qt::UserRole,newItems[i].first);
        ui->lstCogMemory->item(i)->setText(newItems[i].second);
        ui->lstCogMemory->item(i)->setForeground(QColor(255,255,255));
        QColor bgColor(64,64,64);
        if (i == status.programCounter)
            bgColor = QColor(128,0,0);
        else if (m_annotation.annotations[i].stepOver)
            bgColor = QColor(128,128,0);
        ui->lstCogMemory->item(i)->setBackground(bgColor);
        if (resetCurrentIndex && i==status.programCounter)
            ui->lstCogMemory->scrollToItem(ui->lstCogMemory->item(i), QAbstractItemView::EnsureVisible);
    }
    //Update Cog Inner Variables View
    displayInnerVars(m_propeller->listInnerCogVariables(m_cogNo), ui->lstCogInnerVars);


    if (!status.enabled)
        ui->lblStatus->setText(tr("Disabled"));
    else if (!status.running)
        ui->lblStatus->setText(tr("Loading"));
    else {
        QString st = tr("Run, Flg: ");
        st += status.zeroFlag ? "Z" : "-";
        st += status.carryFlag ? "C" : "-";
        st += " "+status.stateName();
        st += tr(" Ctr: 0x%1").arg(m_propeller->globalCounter(),0,16);
        ui->lblStatus->setText(st);
    }
}

QString CogViewer::getTextForCogRamEntry(int address, int data, bool isAtPc) {
    return m_annotation.annotations[address].label.leftJustified(17,' ')+QString::number(address,16).rightJustified(3,'0')+": "+QString::number(quint32(data),16).rightJustified(8,'0')+" "+PDisAsm::disassemble(m_annotation,address,isAtPc,data);
}

void CogViewer::on_lstCogMemory_itemDoubleClicked(QListWidgetItem *item) {
    auto &annotation = m_annotation.annotations[item->data(Qt::UserRole).toInt()];
    auto dlg = new CogRamAnnotationDialog(this, annotation);
    if (dlg->exec() == QDialog::Accepted) {
        annotation = dlg->getResult();
        updateState(false);
        m_annotation.saveFile(m_annotationFileName);
    }
    dlg->deleteLater();
}

void CogViewer::on_btnLoadAnnotation_clicked() {
    m_annotationFileName = QFileDialog::getOpenFileName(this,tr("Load Annotation File"));
    loadAnnotation();
}

void CogViewer::loadAnnotation() {
    if (m_annotationFileName.isEmpty()) {
        m_annotation.clear();
        updateState(false);
        ui->lblAnnotation->setText(tr("No Annotation File loaded"));
        return;
    }
    m_annotation.loadFile(m_annotationFileName);
    updateState(false);
    ui->lblAnnotation->setText(tr("Annotation: %1").arg(m_annotation.name));
}

void CogViewer::displayInnerVars(QList<QPair<QString,QString>> innerVars, QListWidget *lstWdg) {
    for (int i=0; i<innerVars.size(); ++i) {
        QString txt = innerVars[i].first.leftJustified(40,' ')+innerVars[i].second;
        if (i>=lstWdg->count()) {
            lstWdg->addItem("");
            lstWdg->item(i)->setForeground(QColor(255,255,255));
        }

        auto item = lstWdg->item(i);
        if (item->text() != txt) {
            item->setText(txt);
            item->setBackground(QColor(128,0,0));
            item->setData(Qt::UserRole,0);
        }
        else {
            int changedOn = qMin(item->data(Qt::UserRole).toInt(),16);
            item->setData(Qt::UserRole,changedOn+1);
            item->setBackground(QColor(128-changedOn*4,128-changedOn*4,changedOn*4));
        }
    }
}
