#include "i2cdevice.h"
#include <QtDebug>

I2CDevice::~I2CDevice() {
}

I2CEndpoint::I2CEndpoint(QString deviceId, QString deviceName, int sclPin, int sdaPin):
    HWTraceable (deviceId, deviceName),
    m_trace(nullptr),
    m_sclMask(1U<<sclPin),
    m_sdaMask(1U<<sdaPin),
    m_nonSdaMask(0xFFFFFFFF-m_sdaMask),
    m_activeDeviceIndex(-1),
    m_bitCounter(0),
    m_readDataShiftRegister(0),
    m_writeDataShiftRegister(0xFF),
    m_state(Idle),
    m_lastScl(true),
    m_lastSdaIn(true),
    m_sdaOut(true)
{
}

void I2CEndpoint::addDevice(I2CDevice *device) {
    m_devices << device;
}

quint32 I2CEndpoint::clockCycle(quint32 pulledUpPins) {
    //pulledUpPins sind die Pins, bei denen die entsprechenden sda/scl auf 1 gesetzt wurden, sofern diese im tri zustand sind
    //zurückgegeben wird entweder alle bits 1 (wenn sda nicht vom endgerät nach unten gezogen wird) ansonsten ist das entsprechende bit 0
    const bool scl = (pulledUpPins & m_sclMask) != 0;
    const bool sda = (pulledUpPins & m_sdaMask) != 0;

    //


    //wenn scl auf high pegel ist (keine flanke) dann kann mit sda eine start/stop markierung erzeugt werden
    if (m_lastScl && scl) {
        if (m_lastSdaIn && !sda) {
            //Start Condition wurde erkannt
            //eventuell vorhandene kommunikation abbrechen und neu starten
            TRACE(m_trace, "Start", 0, 0);
            stop();
            m_state = Started;
        }
        else if (!m_lastSdaIn && sda) {
            //Stop Condition wurde erkannt
            TRACE(m_trace, "Stop", 0, 0);
            stop();
            m_state = Idle;
        }
    }
    //eine fallende Flanke von scl bedeutet, dass das endgerät jetzt daten setzen muss
    //hier wird auch die statemachine gesteuert
    else if (m_lastScl && !scl) {
        m_bitCounter++;
        if (m_bitCounter==8) { //es wurden 8 bits gelesen und geschrieben, jetzt soll das ack/nack bit geschrieben/gelesen werden
            m_writeDataShiftRegister=0xFF; //nack
            switch (m_state) {
                case Idle: case Started: case StartReadFromDevice: break; //nicht erlaubt
                case ReadAddress: {
                    /* Die I2C Adressse wurde gelesen jetzt muss:
                     * 1. Geprüft werden, ob ein Gerät vorhanden ist
                     * 2. Das ACK/Nack bit gesetzt werden
                     * 3. Je nach unterstem Bit der adresse in den Lese oder Schreibmodus geschaltet werden
                     */
                    for(int d=0; d<m_devices.size(); ++d) {
                        if (!m_devices[d]->start(m_readDataShiftRegister))
                            continue;
                        if (m_activeDeviceIndex>=0) {
                            qWarning()<<"multiple devices responding to address"<<m_readDataShiftRegister;
                            m_devices[m_activeDeviceIndex]->stop();
                            m_activeDeviceIndex=-1;
                            break;
                        }
                        m_activeDeviceIndex = d;
                    }
                    if (m_activeDeviceIndex>=0) {
                        m_writeDataShiftRegister=0; //ack senden, wenn gerät gefunden
                        if ((m_readDataShiftRegister&1) == 0) //es wird geschrieben
                            m_state = WriteToDevice;
                        else
                            m_state = StartReadFromDevice;
                    }
                    else {
                        qWarning()<<"no device found with address"<<m_readDataShiftRegister;
                        m_state = Cancelled;
                    }
                    break;
                }
                case WriteToDevice:
                    /*
                     * Es wurde ein Byte empfangen, das an das Gerät geschickt werden soll
                     * Es muss jetzt das ACK/NACK Bit gesetzt werden
                     */
                    if (m_devices[m_activeDeviceIndex]->writeToDevice(m_readDataShiftRegister))
                        m_writeDataShiftRegister=0;
                    else
                        m_state = Cancelled;
                    break;
                case ReadFromDevice:
                    /*
                     * Das Byte vom Gerät wurde zwar schon übertragen, aber noch nicht das ACK/NACK
                     * Bit vom Master empfangen, deshalb hier nichts machen
                     */
                    break;
                case Cancelled:
                    //Die Kommunikation ist bereits abgebrochen worden
                    break;

            }
            m_readDataShiftRegister = 0;
        }
        else if (m_bitCounter==9) { //das ack/nack bit wurde gelesen/geschrieben, jetzt beginnt ein neues frame
            m_bitCounter = 0;
            m_writeDataShiftRegister = 0xFF;
            switch(m_state) {
                case Idle: case ReadAddress: break;//das ist nicht erlaubt
                case Cancelled:
                    //Die Kommunikation ist bereits abgebrochen worden
                    break;
                case Started: //eben wurde der start marker gesendet, es folgt die addresse
                    m_state = ReadAddress;
                    break;
                case WriteToDevice:
                    //Es muss nichts an den Master geschrieben werden
                    break;
                case StartReadFromDevice:
                    //Die Adresse wurde verarbeitet, es wird das erste Byte gelesen
                    m_writeDataShiftRegister = m_devices[m_activeDeviceIndex]->readFromDevice(true); //ack
                    m_state = ReadFromDevice;
                    break;
                case ReadFromDevice:
                    //Das nächste Byte wird gelesen
                    m_writeDataShiftRegister = m_devices[m_activeDeviceIndex]->readFromDevice(m_readDataShiftRegister==0);
                    break;
            }
            m_readDataShiftRegister = 0;
        }
        m_sdaOut = (m_writeDataShiftRegister&0x80) != 0;
        m_writeDataShiftRegister<<=1;
    }
    //eine steigende Flanke von scl bedeutet, dass daten vom master gelesen werden können
    else if (!m_lastScl && scl) {
        //steigende flanke, hier werden daten gelesen
        m_readDataShiftRegister = (m_readDataShiftRegister<<1) | int(sda);
    }
    m_lastScl = scl;
    m_lastSdaIn = sda;
    return m_nonSdaMask | (m_sdaOut ? m_sdaMask : 0UL);
}

void I2CEndpoint::stop() {
    m_bitCounter = 8;
    m_sdaOut = true;
    m_readDataShiftRegister = 0;
    m_writeDataShiftRegister = 0xFF;
    if (m_activeDeviceIndex>=0)
        m_devices[m_activeDeviceIndex]->stop();
    m_activeDeviceIndex = -1;
}

TraceInfo I2CEndpoint::getDeviceTraceInfo() const {
    return HWTraceable::getDeviceTraceInfo().addMethod("I2CBase");
}

void I2CEndpoint::setDebugTraceFunction(QString method, HWTraceMethod func) {
    if (method == "I2CBase")
        m_trace = func;
}

void I2CEndpoint::serialize(QDataStream& stream) const {
    HWTraceable::serialize(stream);
    stream << m_activeDeviceIndex;
    stream << m_bitCounter;
    stream << m_readDataShiftRegister;
    stream << m_writeDataShiftRegister;
    stream << m_state;
    stream << m_lastScl;
    stream << m_lastSdaIn;
    stream << m_sdaOut;
}

void I2CEndpoint::deserialize(QDataStream& stream) {
    HWTraceable::deserialize(stream);

    stream >> m_activeDeviceIndex;
    stream >> m_bitCounter;
    stream >> m_readDataShiftRegister;
    stream >> m_writeDataShiftRegister;

    int tmpState=0;
    stream >> tmpState;
    m_state = static_cast<I2CState>(tmpState);

    stream >> m_lastScl;
    stream >> m_lastSdaIn;
    stream >> m_sdaOut;
}

