#include "i2ceeprom.h"
#include <QtDebug>

I2CEEPRom::I2CEEPRom(QString deviceId, QString deviceName, QByteArray initialData, int size):
    HWTraceable(deviceId, deviceName),
    m_trace(nullptr),
    m_state(Idle),
    m_dataPointer(0),
    m_storedData(initialData)
{
    m_storedData.resize(size);
}

bool I2CEEPRom::start(int address) {
    m_state = Idle;
    if (address == 160) { //schreiben
        m_state = AddrHigh;
        return true;
    }
    if (address == 161) { //lesen
        m_state = Read;
        return true;
    }
    return false;
}

void I2CEEPRom::stop() {
    m_state = Idle;
}

bool I2CEEPRom::writeToDevice(int data) {
    switch(m_state) {
        case Idle: case Read: m_state = Idle; return false; //illegal
        case AddrHigh:
            TRACE(m_trace,"SetAddrPtrHigh",data,0);
            m_dataPointer = data<<8;
            m_state = AddrLow;
            return true;
        case AddrLow:
            TRACE(m_trace,"SetAddrPtrLow",data,0);
            m_dataPointer += data;
            m_state = Write;
            return true;
        case Write:
            TRACE(m_trace,"EEPROM WriteData",m_dataPointer,data);
            qDebug()<<"eeprom write not supported"<<data;
            m_dataPointer++;
            return false;
    }
    return false;
}

int I2CEEPRom::readFromDevice(bool lastByteAck) {
    switch(m_state) {
        case Idle: case Write: case AddrLow: case AddrHigh: m_state = Idle; return 0xFF; //illegal
        case Read: {
            if (!lastByteAck || m_dataPointer>=m_storedData.size()) {
                TRACE(m_trace,"EEPROM ReadStopped",0,0);
                m_state = Idle;
                return 0xFF;
            }
            const int result = int(quint8(m_storedData[m_dataPointer]));
            TRACE(m_trace,"EEPROM ReadByte",m_dataPointer,result);
            m_dataPointer++;
            return result;
        }
    }
    return 0xFF;
}

TraceInfo I2CEEPRom::getDeviceTraceInfo() const {
    return HWTraceable::getDeviceTraceInfo().addMethod("I2CEEPROM");
}

void I2CEEPRom::setDebugTraceFunction(QString method, HWTraceMethod func) {
    if (method == "I2CEEPROM")
        m_trace = func;
}

void I2CEEPRom::serialize(QDataStream& stream) const {
    HWTraceable::serialize(stream);
    stream << m_state;
    stream << m_dataPointer;
    stream << m_storedData;
}

void I2CEEPRom::deserialize(QDataStream& stream) {
    HWTraceable::deserialize(stream);
    int tmpState = 0;
    stream >> tmpState;
    m_state = static_cast<State>(tmpState);
    stream >> m_dataPointer;
    stream >> m_storedData;
}
