#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "propellerwidget.h"
#include <QtDebug>
#include <QTime>
#include <QMessageBox>
#include <QInputDialog>

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    m_sim(new SimulationMaster(SimulationMaster::PropNewCpp)), //PropNewCpp
    m_simGolden(nullptr),
    //m_simGolden(new SimulationMaster(SimulationMaster::PropGenCpp)), //expected golden simulation
    m_simCompareWidget(nullptr),
    m_breakpointReached(false)
{
    ui->setupUi(this);
    setRunning(false);
    ui->wdgHive->setSimulationMaster(m_sim);
    ui->wdgCog->setPropeller(m_sim->propeller(), 0);
    ui->wdgHub->setPropeller(m_sim->propeller());
    setWindowTitle(m_sim->propeller()->name());

    if (m_simGolden) {
        m_simCompareWidget = new PropellerWidget();
        m_simCompareWidget->show();
        m_simCompareWidget->setPropeller(m_simGolden->propeller());
    }

    //on_actionLoad_Snapshot_triggered();

    /*
    if (m_sim->propeller()->name() == "PropellerNewCpp")
        loadSnapshot("other-files/snapshot.bin.new");
    else
        loadSnapshot("other-files/snapshot.bin.old");*/
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::on_spbCogNo_valueChanged(int) {
    ui->wdgCog->setPropeller(m_sim->propeller(), ui->spbCogNo->value());
}

void MainWindow::on_actionSingleClockCycle_triggered() {
    for (int i=0; i<1; ++i) {
        m_sim->clockCycle();
        if (m_simGolden)
            m_simGolden->clockCycle();
    }
    updateViews();
}

void MainWindow::on_actionRun_until_global_counter_triggered() {
    bool ok = false;
    const quint32 destCounter = QInputDialog::getText(this,"Run until","Run until",QLineEdit::Normal,QString::number(m_sim->propeller()->globalCounter())).toUInt(&ok);
    if (!ok)
        return;

    int simCompareCounter=0;
    while(m_sim->propeller()->globalCounter() != destCounter) {
        m_sim->clockCycle();
        if (m_simGolden) {
            m_simGolden->clockCycle();
            ++simCompareCounter;
            if (simCompareCounter>100) {
                simCompareCounter=0;
                m_sim->compareWith(m_simGolden);
            }
        }
    }
    updateViews();
}

void MainWindow::on_actionNextAsmInstruction_triggered() {
    const int selCog = ui->spbCogNo->value();
    auto *prop = m_sim->propeller();
    const int origPC = prop->cogStatus(selCog).programCounter;
    auto stepOverList = ui->wdgCog->getStepOverList();
    setupDebugTracer([this,prop,selCog,origPC,stepOverList](QString deviceId, QString methodName) {
        if (deviceId == "M.P" && methodName == "ClockCycle") {
            return HWTraceMethod([this,prop,selCog,origPC,stepOverList](const char*, int, int) {
                const auto status = prop->cogStatus(selCog);
                if (status.running && status.enabled && status.programCounter != origPC && !stepOverList[status.programCounter])
                    triggerBreakpoint();
            });
        }
        if (deviceId == "M.I" && methodName == "I2CBase") {
            return HWTraceMethod([](const char* action, int d1, int d2) {
                qDebug()<<action<<d1<<d2;
            });

        }
        if (deviceId == "M.I.E" && methodName == "I2CEEPROM") {
            return HWTraceMethod([](const char* action, int d1, int d2) {
                qDebug()<<action<<d1<<d2;
            });

        }
        return HWTraceMethod();
    });
    runUntilBreakpoint();
}

void MainWindow::updateViews() {
    setWindowTitle(QString("%1: %2").arg(m_sim->propeller()->name()).arg(m_sim->propeller()->globalCounter()));
    if (m_simGolden)
        m_sim->compareWith(m_simGolden);
    ui->wdgCog->updateState(true);
    ui->wdgHub->updateState();
    if (m_simCompareWidget)
        m_simCompareWidget->updateViews();
    QString pinOutput;
    const quint32 pinDir = m_sim->propeller()->pinDir();
    const quint32 pinOut = m_sim->propeller()->pinOut();
    for (int i=31; i>=0; --i) {
        pinOutput += (pinDir & (1<<i)) == 0 ? "T" : ((pinOut & (1<<i)) != 0 ? "1" : "0");
        if (i%8==0)
            pinOutput += " ";
    }
    ui->lblPinOutput->setText(pinOutput);
}

void MainWindow::setupDebugTracer(std::function<HWTraceMethod(QString deviceId, QString methodName)> callback) {
    foreach(auto dev, m_sim->getDevices()) {
        auto trc = dev->getDeviceTraceInfo();
        foreach (auto methodName, trc.traceMethods)
            dev->setDebugTraceFunction(methodName, callback(trc.deviceId, methodName));
    }
}

HWTraceMethod MainWindow::generateDebugMethod(QString deviceId) {
    return [deviceId](const char*method, int data1, int data2) {
        qDebug()<<deviceId<<method<<data1<<data2;
    };
}

void MainWindow::triggerBreakpoint() {
    m_breakpointReached = true;
}

void MainWindow::setRunning(bool running) {
    ui->actionSingleClockCycle->setEnabled(!running);
    ui->actionNextAsmInstruction->setEnabled(!running);
    ui->actionRunUntilBreakpoint->setEnabled(!running);
    ui->actionSave_Snapshot->setEnabled(!running);
    ui->actionLoad_Snapshot->setEnabled(!running);

    ui->actionBreak->setEnabled(running);
}

void MainWindow::runUntilBreakpoint() {
    setRunning(true);
    m_breakpointReached = false;
    int cycleCounter = 0;
    QTime start;
    start.start();

    const int processEventsEach = 200000;
    //const int processEventsEach = 2000;

    int simCompareCounter=0;

    while (!m_breakpointReached) {
        m_sim->clockCycle();
        if (m_simGolden) {
            m_simGolden->clockCycle();
            ++simCompareCounter;
            if (simCompareCounter>100) {
                simCompareCounter=0;
                m_sim->compareWith(m_simGolden);
            }
        }
        ++cycleCounter;
        if (cycleCounter==processEventsEach) {
            QApplication::processEvents();
            int elapsed = start.elapsed();
            if (elapsed == 0)
                setWindowTitle(tr("Simulation Frequency: inf kHz"));
            else if (cycleCounter>elapsed*10)
                setWindowTitle(tr("Simulation Frequency: %1 kHz").arg(cycleCounter/elapsed));
            else
                setWindowTitle(tr("Simulation Frequency: %1 Hz").arg(cycleCounter*1000/elapsed));
            cycleCounter = 0;
            start.start();
        }
    }
    setRunning(false);
    updateViews();
}

void MainWindow::on_actionRunUntilBreakpoint_triggered() {
    setupDebugTracer([this](QString deviceId, QString methodName) {
        if (deviceId == "M.I.E" && methodName == "I2CEEPROM") {
            return HWTraceMethod([this](const char*, int d1, int) {
                //qDebug()<<"TRCE"<<d1;
                if (d1%1000==0 || d1>= 32767)
                    qDebug()<<d1;
                if (d1 == 32768)
                    triggerBreakpoint();
            });
        }
        return HWTraceMethod();
    });
    runUntilBreakpoint();
}

void MainWindow::on_actionBreak_triggered() {
    triggerBreakpoint();
}

void MainWindow::loadSnapshot(QString file) {
    QByteArray snap = FileTools::readBinary(file);
    if (snap.isEmpty())
        return;
    try {
        QDataStream stream(snap);
        m_sim->deserialize(stream);
        if (m_simGolden)
            m_simGolden->deserialize(stream);
        updateViews();
    }
    catch(HWSerializeException& e) {
        QMessageBox::warning(this,tr("Unable to load snapshot"),tr("Error while loading snapshot: %1").arg(e.message));
        return;
    }
}

void MainWindow::on_actionLoad_Snapshot_triggered() {
    loadSnapshot("other-files/snapshot.bin");
}

void MainWindow::on_actionSave_Snapshot_triggered() {
    try {
        QByteArray snapShot;
        QDataStream stream(&snapShot,QIODevice::WriteOnly);
        m_sim->serialize(stream);
        if (m_simGolden)
            m_simGolden->serialize(stream);
        if (!stream.atEnd())
            throw HWSerializeException("expected end of stream");
        FileTools::saveBinary("other-files/snapshot.bin", snapShot);
    }
    catch(HWSerializeException& e) {
        QMessageBox::warning(this,tr("Unable to save snapshot"),tr("Error while saving snapshot: %1").arg(e.message));
        return;
    }
}
