#ifndef PROPELLERASMMEMORYMAP_H
#define PROPELLERASMMEMORYMAP_H

#include <QtDebug>
#include <QList>
#include <QString>
#include <QPair>
#include <QDataStream>
#include <QSharedMemory>

namespace PropellerAsmInternals {
    class MemoryMap {
    public:
        enum {
            CogMemoryMultiplicator      = 1024*4,
            //Offset beziehen sich auf die MemoryMap
            StartOfDataMemoryMap        = 0,
            GlobalHubMemoryOffset       = StartOfDataMemoryMap,                             HubMemorySize       = 1024*96,
            GlobalCogMemoryOffset       = GlobalHubMemoryOffset+HubMemorySize,              CogMemorySize       = CogMemoryMultiplicator*8,
            GlobalCogRegisterOffset     = GlobalCogMemoryOffset+CogMemorySize,              CogRegisterSize     = CogMemoryMultiplicator*8,
            GlobalHubRegisterOffset     = GlobalCogRegisterOffset+CogRegisterSize,          HubRegisterSize     = 1024*4,
            GlobalScratchpadOffset      = GlobalHubRegisterOffset+HubRegisterSize,          ScratchPadSize      = 1024*4,
            EndOfDataMemoryMap          = GlobalScratchpadOffset+ScratchPadSize,

            StartOfCodeMemoryMap        = EndOfDataMemoryMap,
            GlobalPreEmptyAreaOffset    = StartOfCodeMemoryMap,                             PreEmptyAreaSize    = 1024*4,
            GlobalDynamicCogAsmOffset   = GlobalPreEmptyAreaOffset+PreEmptyAreaSize,        DynamicCogAsmOffset = CogMemoryMultiplicator*8,
            GlobalPostEmptyAreaOffset   = GlobalDynamicCogAsmOffset+DynamicCogAsmOffset,    PostEmptyAreaSize   = 1024*4,
            EndOfCodeMemoryMap          = GlobalPostEmptyAreaOffset+PostEmptyAreaSize,

            TotalDataMemoryMapSize      = EndOfDataMemoryMap-StartOfDataMemoryMap,
            TotalCodeMemoryMapSize      = EndOfCodeMemoryMap-StartOfCodeMemoryMap,
            TotalMemoryMapSize          = TotalDataMemoryMapSize+TotalCodeMemoryMapSize,

            //Relativ zum GlobalHubRegisterOffset, darauf zeigt der hubPtr
            HubRelMemoryOffset          = GlobalHubMemoryOffset-GlobalHubRegisterOffset,
            HubRelScratchpadOffset      = GlobalScratchpadOffset-GlobalHubRegisterOffset,

            //Relativ zum GlobalCogRegisterOffset, darauf (+cogNo*1024*4) zeigt der cogPtr
            CogRelMemoryOffset          = GlobalCogMemoryOffset-GlobalCogRegisterOffset,
            CogRelDynamicAsmOffset      = GlobalDynamicCogAsmOffset-GlobalCogRegisterOffset,

            //Interne Werte
            InternalRawMemorySize = TotalMemoryMapSize+0x2000 //zwei Pages größer reservieren, damit noch ausgerichtet werden kann
        };
        //TODO auch ohne sharedMem anbieten
        MemoryMap(QString sharedMemKey, bool isMaster):
            m_sharedMem(sharedMemKey.isEmpty() ? nullptr : new QSharedMemory(sharedMemKey)),
            m_nonSharedMem(sharedMemKey.isEmpty() ? new quint8[InternalRawMemorySize] : nullptr),
            m_startOfMemoryMap(nullptr),
            m_masterPointer(0)
        {
            quint8 *ptr = m_nonSharedMem;
            if (m_sharedMem) {
                if (m_sharedMem->attach()) {
                    qDebug()<<"Attached to shared Memory"<<sharedMemKey;
                    if (m_sharedMem->size()<InternalRawMemorySize) {
                        m_sharedMem->detach();
                        throw QString("shared memory too small");
                    }
                }
                else if (!m_sharedMem->create(InternalRawMemorySize))
                    throw QString("unable to create shared memory");
                ptr = reinterpret_cast<quint8*>(m_sharedMem->data());
            }
            if (!ptr)
                throw QString("raw mem ptr was null");
            if (isMaster) {
                //Der Master richtet sich an der Page grenze aus
                memset(ptr,0,InternalRawMemorySize); //alles löschen
                m_startOfMemoryMap = reinterpret_cast<quint8*>((reinterpret_cast<quint64>(ptr)+0x1000ULL) & (~0xFFFULL));
                auto sigCpy = signature();
                memcpy(m_startOfMemoryMap-16,sigCpy.constData(),8);
                memcpy(m_startOfMemoryMap-8,&m_startOfMemoryMap,8); //echten pointer kopieren
            }
            else
                redoAlignmentSearch();
        }
        ~MemoryMap() {
            delete m_sharedMem;
        }
        static QByteArray signature() { return "PropAsmM"; }
        void redoAlignmentSearch() {
            if (!m_sharedMem)
                throw QString("redoAlignment only allowed with shared memory");
            m_masterPointer = 0;
            quint8* ptr = reinterpret_cast<quint8*>(m_sharedMem->data());
            m_startOfMemoryMap = reinterpret_cast<quint8*>((reinterpret_cast<quint64>(ptr)+0x1000ULL) & (~0xFFFULL));
            //nur im client debugger, nie im eigentlichen propeller
            int j=0;
            while(j<InternalRawMemorySize && ptr[j] == 0)
                ++j;
            //Signatur vergleichen
            for (int i=0; i<signature().size(); ++i, ++j) {
                if (j>=InternalRawMemorySize || ptr[j] != quint8(signature()[i])) {
                    qWarning()<<"signature not found in shared memory";
                    return;
                }
            }
            memcpy(&m_masterPointer,&ptr[j],8);
            qDebug()<<"found signature"<<m_masterPointer;
            j+=8;
            if (InternalRawMemorySize-j<TotalMemoryMapSize) {
                qWarning()<<"remaining shared memory after signature too small";
                return;
            }
            m_startOfMemoryMap = &ptr[j];
        }
        quint64 getMasterPointer() const { return m_masterPointer; }
        quint8* fullMemoryMap() const { return m_startOfMemoryMap; }
        quint8* dataMemoryMap() const { return &m_startOfMemoryMap[StartOfDataMemoryMap]; }
        quint8* codeMemoryMap() const { return &m_startOfMemoryMap[StartOfCodeMemoryMap]; }
        quint8* hubMemory() const { return &m_startOfMemoryMap[GlobalHubMemoryOffset]; }
        quint8* hubRegisters() const { return &m_startOfMemoryMap[GlobalHubRegisterOffset]; }
        quint8* scratchPad() const { return &m_startOfMemoryMap[GlobalScratchpadOffset]; }
        quint8* cogMemory(int cog) const { return &m_startOfMemoryMap[GlobalCogMemoryOffset+CogMemoryMultiplicator*cog]; }
        quint8* cogRegisters(int cog) const { return &m_startOfMemoryMap[GlobalCogRegisterOffset+CogMemoryMultiplicator*cog]; }
        quint8* cogDynamicAsm(int cog) const { return &m_startOfMemoryMap[GlobalDynamicCogAsmOffset+CogMemoryMultiplicator*cog]; }
    private:
        QSharedMemory *m_sharedMem;
        quint8* m_nonSharedMem;
        quint8* m_startOfMemoryMap;
        quint64 m_masterPointer;
    };

    struct AsmFunctionEntry {
        enum Type { FunctionStatic=0,FunctionDynamicHub=1,FunctionDynamicCog=2,Template=3 };
        AsmFunctionEntry():type(FunctionStatic),relOffset(0),size(0) {}
        AsmFunctionEntry(QString name, Type type, int relOffset, int size):name(name),type(type),relOffset(relOffset),size(size) {}
        QString name;
        Type type;
        int relOffset; //zu der entsprechenden sektion
        int size;
    };
}

#endif // PROPELLERASMMEMORYMAP_H
