#include "ps2device.h"
#include <QtDebug>

PS2Device::PS2Device(quint32 clockPin, quint32 dataPin):
    m_clockMask(1U<<clockPin),
    m_dataMask(1U<<dataPin),
    m_remainingMask(~(m_clockMask|m_dataMask)),
    m_clockDivider(0),
    m_pinOut(0xFFFFFFFFU),
    m_bitCounter(0),
    m_recv(0),
    m_send(0),
    m_packageSendPos(0),
    m_state(Idle)
{

}

PS2Device::~PS2Device() {
}

quint32 PS2Device::clockCycle(quint32 pulledUpPins) {
    m_clockDivider++;
    if (m_clockDivider<4000)
        return m_pinOut;
    m_clockDivider = 0;

    bool clk = (pulledUpPins&m_clockMask) != 0U;
    bool data = (pulledUpPins&m_dataMask) != 0U;

    //Wenn Host clk auf low zieht, dann fordert er ein host->device transfer an
    //das geht zu jeder Zeit
    if (m_state != HostWaitStart && !clk) {
        qDebug()<<"Host requested transmission (clk low)";
        m_packageSendPos = 0;
        m_state = HostWaitStart;
        return m_pinOut;
    }
    if (m_state == HostWaitStart && !data && clk) {
        m_state = HostToDeviceClkLow;
        m_bitCounter = 0;
        m_recv = 0;
        return m_pinOut;
    }
    if (m_state == HostToDeviceClkLow) {
        m_pinOut = m_remainingMask | m_dataMask;
        m_state = HostToDeviceClkHigh;
        return m_pinOut;
    }

    if (m_state == HostToDeviceClkHigh) {
        m_pinOut = m_remainingMask | m_dataMask | m_clockMask;
        m_state = HostToDeviceClkLow;
        //qDebug()<<"CLK IN"<<m_bitCounter<<data;
        if (data)
            m_recv |= (1<<m_bitCounter);
        m_bitCounter++;

        if (m_bitCounter == 10)
            m_state = HostToDeviceStop1;

        return m_pinOut;
    }

    if (m_state == HostToDeviceStop1) {
        m_pinOut = m_remainingMask | m_clockMask;
        m_state = HostToDeviceStop2;
        return m_pinOut;
    }

    if (m_state == HostToDeviceStop2) {
        m_pinOut = m_remainingMask;
        m_state = HostToDeviceStop3;
        return m_pinOut;
    }

    if (m_state == HostToDeviceStop3) {
        if (hasOddBits(m_recv&0x1FF)) {
            qDebug()<<"recv byte"<<(m_recv&0xFF);
            m_commandResponse = byteReceived(m_recv&0xFF);
        }
        else {
            qWarning()<<"parity error";
        }

        m_pinOut = m_remainingMask | m_dataMask | m_clockMask;
        m_state = Idle;
        return m_pinOut;
    }
    if (m_state == Idle && !m_commandResponse.isEmpty()) {
        transmitByte(int(quint8(m_commandResponse[0])));
        m_packageSendPos = 0;
        return m_pinOut;
    }
    if (m_state == Idle && !m_packages.isEmpty()) {
        removeFinishedPackages();
        if (!m_packages.isEmpty()) {
            transmitByte(int(quint8(m_packages[0][m_packageSendPos])));
            return m_pinOut;
        }
    }
    if (m_state == DeviceToHostClockLow) {
        //qDebug()<<"SEND BIT"<<m_bitCounter<<((m_send&1)==1);
        m_pinOut = m_remainingMask | ((m_send&1)==1 ? m_dataMask : 0);
        m_send >>= 1;
        m_state = DeviceToHostClockHigh;
        m_bitCounter++;
        return m_pinOut;
    }
    if (m_state == DeviceToHostClockHigh) {
        m_pinOut |= m_clockMask;
        m_state = DeviceToHostClockLow;
        if (m_bitCounter == 11) {
            if (!m_commandResponse.isEmpty()) //war eine cmd response
                m_commandResponse.remove(0,1);
            else { //war device package data
                m_packageSendPos++;
                removeFinishedPackages();
            }
            m_state = Idle;
        }
    }
    return m_pinOut;
}

void PS2Device::removeFinishedPackages() {
    while (!m_packages.isEmpty() && m_packageSendPos >= m_packages.first().size()) {
        m_packages.removeAt(0);
        m_packageSendPos = 0;
    }
}

void PS2Device::transmitByte(int b) {
    qDebug()<<"send byte"<<b;
    m_send = b<<1;
    if (!hasOddBits(m_send))
        m_send |= 0x200;
    m_send |= 0x400;
    m_bitCounter = 0;
    m_state = DeviceToHostClockLow;
}

bool PS2Device::hasOddBits(int data) const {
    bool parity = false;
    for (int i=0; i<32; ++i) {
        if ((data&1) == 1)
            parity = !parity;
        data >>= 1;
    }
    return parity;
}

void PS2Device::transmitPackage(QByteArray pkg) {
    m_packages << pkg;
}

PS2Keyboard::PS2Keyboard(quint32 clockPin, quint32 dataPin):
    PS2Device (clockPin, dataPin),
    m_kbdState(KbdIdle)
{
    m_keyMap[Qt::Key_A] = ScanCode(0x1C);
    m_keyMap[Qt::Key_B] = ScanCode(0x32);
    m_keyMap[Qt::Key_C] = ScanCode(0x21);
    m_keyMap[Qt::Key_D] = ScanCode(0x23);
    m_keyMap[Qt::Key_E] = ScanCode(0x24);
    m_keyMap[Qt::Key_F] = ScanCode(0x2B);
    m_keyMap[Qt::Key_G] = ScanCode(0x34);
    m_keyMap[Qt::Key_H] = ScanCode(0x33);
    m_keyMap[Qt::Key_I] = ScanCode(0x42);
    m_keyMap[Qt::Key_J] = ScanCode(0x3B);
    m_keyMap[Qt::Key_K] = ScanCode(0x42);
    m_keyMap[Qt::Key_L] = ScanCode(0x4B);
    m_keyMap[Qt::Key_M] = ScanCode(0x3A);
    m_keyMap[Qt::Key_N] = ScanCode(0x31);
    m_keyMap[Qt::Key_O] = ScanCode(0x44);
    m_keyMap[Qt::Key_P] = ScanCode(0x4D);
    m_keyMap[Qt::Key_Q] = ScanCode(0x15);
    m_keyMap[Qt::Key_R] = ScanCode(0x2D);
    m_keyMap[Qt::Key_S] = ScanCode(0x1B);
    m_keyMap[Qt::Key_T] = ScanCode(0x2C);
    m_keyMap[Qt::Key_U] = ScanCode(0x3C);
    m_keyMap[Qt::Key_V] = ScanCode(0x2A);
    m_keyMap[Qt::Key_W] = ScanCode(0x1D);
    m_keyMap[Qt::Key_X] = ScanCode(0x22);
    m_keyMap[Qt::Key_Y] = ScanCode(0x35);
    m_keyMap[Qt::Key_Z] = ScanCode(0x1A);

    m_keyMap[Qt::Key_0] = ScanCode(0x45);
    m_keyMap[Qt::Key_1] = ScanCode(0x16);
    m_keyMap[Qt::Key_2] = ScanCode(0x1E);
    m_keyMap[Qt::Key_3] = ScanCode(0x26);
    m_keyMap[Qt::Key_4] = ScanCode(0x25);
    m_keyMap[Qt::Key_5] = ScanCode(0x2E);
    m_keyMap[Qt::Key_6] = ScanCode(0x36);
    m_keyMap[Qt::Key_7] = ScanCode(0x3D);
    m_keyMap[Qt::Key_8] = ScanCode(0x3E);
    m_keyMap[Qt::Key_9] = ScanCode(0x46);
    m_keyMap[Qt::Key_Backspace] = ScanCode(0x66);
    m_keyMap[Qt::Key_Space] = ScanCode(0x29);
    m_keyMap[Qt::Key_Tab] = ScanCode(0x0D);
    m_keyMap[Qt::Key_Shift] = ScanCode(0x12);
    m_keyMap[Qt::Key_Control] = ScanCode(0x14);
    m_keyMap[Qt::Key_Alt] = ScanCode(0x11);
    m_keyMap[Qt::Key_Enter] = ScanCode(0x5A);
    m_keyMap[Qt::Key_Escape] = ScanCode(0x76);

    m_keyMap[Qt::Key_F1] = ScanCode(0x05);
    m_keyMap[Qt::Key_F2] = ScanCode(0x06);
    m_keyMap[Qt::Key_F3] = ScanCode(0x04);
    m_keyMap[Qt::Key_F4] = ScanCode(0x0C);
    m_keyMap[Qt::Key_F5] = ScanCode(0x03);
    m_keyMap[Qt::Key_F6] = ScanCode(0x0B);
    m_keyMap[Qt::Key_F7] = ScanCode(0x83);
    m_keyMap[Qt::Key_F8] = ScanCode(0x0A);
    m_keyMap[Qt::Key_F9] = ScanCode(0x01);
    m_keyMap[Qt::Key_F10] = ScanCode(0x09);
    m_keyMap[Qt::Key_F11] = ScanCode(0x78);
    m_keyMap[Qt::Key_F12] = ScanCode(0x07);

    m_keyMap[Qt::Key_Period] = ScanCode(0x49);
    m_keyMap[Qt::Key_Comma] = ScanCode(0x41);
    m_keyMap[Qt::Key_Semicolon] = ScanCode(0x4C);
    //m_keyMap[Qt::Key_Equal] = ScanCode(0x55);
    //m_keyMap[Qt::Key_Minus] = ScanCode(0x4E);

}

QByteArray PS2Keyboard::byteReceived(int byte) {
    QByteArray response(1,char(0xFA));
    if (byte == 0xFF) {
        m_kbdState = KbdIdle;
        response.append(char(0xAA));
        return response;
    }
    if (byte == 0xF3) {
        m_kbdState = KbdSetTypematicRateDelay;
        return response;
    }
    if (byte == 0xED) {
        m_kbdState = KbdSetLeds;
        return response;
    }
    if (m_kbdState == KbdSetTypematicRateDelay) {
        qDebug()<<"Set Typematic Rate Delay"<<byte;
        m_kbdState = KbdIdle;
        return response;
    }
    if (m_kbdState == KbdSetLeds) {
        qDebug()<<"Set Leds"<<byte;
        m_kbdState = KbdIdle;
        return response;
    }
    return QByteArray();
}

void PS2Keyboard::keyAction(int key, bool press) {
    ScanCode sc = m_keyMap.value(key, ScanCode());
    if (sc.c1<0) {
        qWarning()<<"unknown key"<<key;
        return;
    }
    QByteArray pkg;
    if (!press)
        pkg.append(char(0xF0));
    pkg.append(char(sc.c1));
    transmitPackage(pkg);
}
