#include "vgadecoder.h"
#include <QtDebug>

VGADecoder::VGADecoder(QString deviceId, QString deviceName, quint32 firstPinNumber):
    HWTraceable (deviceId, deviceName),
    m_colorTable(256),
    m_timeCode(0),
    m_firstPinNumber(firstPinNumber),
    m_vSyncPinMask(0x01u<<firstPinNumber),
    m_hSyncPinMask(0x02u<<firstPinNumber),
    m_lastHsyncTimeStamp(0U),
    m_lastvVsyncLineCounter(0U),
    m_sumHsyncLow(0U),
    m_sumHsyncHigh(0U),
    m_sumVsyncLow(0U),
    m_frameCounter(0U),
    m_hNominator(0U),
    m_hDenominator(0U),
    m_hFrac(0U),
    m_x(0),
    m_y(0),
    m_vFrontPorch(0),
    m_hFrontPorch(0),
    m_lastVsync(false),
    m_lastHsync(false),
    m_frozenVSync(false),
    m_hSyncPolarity(false),
    m_vSyncPolarity(false)
{
    for (int i=0; i<m_colorTable.size(); ++i)
        m_colorTable[i] = qRgb(((i>>4)&3)*85, ((i>>2)&3)*85, (i&3)*85);
}

void VGADecoder::clockCycle(quint32 pins) {
    m_timeCode++;
    const bool hsync = m_hSyncPolarity ^ ((pins & m_hSyncPinMask) != 0);
    if (hsync != m_lastHsync) {
        const quint32 diff = m_timeCode-m_lastHsyncTimeStamp;
        m_lastHsyncTimeStamp = m_timeCode;
        m_lastHsync = hsync;
        if (hsync) { //risign edge
            m_lastvVsyncLineCounter++;
            m_sumHsyncLow += diff;
            m_x = m_hFrontPorch;
            m_hFrac = 0;
            m_frozenVSync = m_lastVsync;
        }
        else {
            m_sumHsyncHigh += diff;
            m_y++;
        }
    }
    const bool vsync = m_vSyncPolarity ^ ((pins & m_vSyncPinMask) != 0);
    if (vsync != m_lastVsync) {
        if (vsync) {
            m_sumVsyncLow = m_lastvVsyncLineCounter;
            m_y = m_vFrontPorch;
        }
        else {
            if (!m_nextImage.isNull()) {
                qDebug()<<"VGA FRAME finished";
                emit nextFrame(m_nextImage);
                //m_nextImage.save(QString("frame-%1.png").arg(m_frameCounter));
            }
            detectFrameFormat();
            m_sumVsyncLow = 0;
            m_sumHsyncLow = 0;
            m_sumHsyncHigh = 0;
            m_frameCounter++;
        }
        m_lastVsync = vsync;
        m_lastvVsyncLineCounter = 0;
    }

    if (hsync && m_frozenVSync && !m_nextImage.isNull() && m_y>=0 && m_y<m_nextImage.height()) {
        m_hFrac += m_hNominator;
        if (m_hFrac >= m_hDenominator) {
            m_hFrac -= m_hDenominator;
            m_x += 1;

            if (m_x>=0 && m_x<m_nextImage.width())
                m_nextImage.scanLine(m_y)[m_x] = (pins >> (m_firstPinNumber+2))&0x3F;
        }
    }

    //qDebug()<<"VGA PINS"<<m_timeCode<<QString::number(pins>>m_firstPinNumber,2).rightJustified(8,'0');
}

void VGADecoder::detectFrameFormat() {
    if (m_sumVsyncLow > m_lastvVsyncLineCounter) {
        m_vSyncPolarity = !m_vSyncPolarity;
        qDebug()<<"Swap VSYNC Polarity"<<m_vSyncPolarity;
        return;
    }
    const quint32 totalLines = m_lastvVsyncLineCounter+m_sumVsyncLow;
    if (totalLines==0 || totalLines>4096) {
        qDebug()<<"Invalid Frame Format VSYNC Total Lines"<<totalLines;
        m_nextImage = QImage();
        return;
    }
    const quint32 avgHsyncLow = m_sumHsyncLow/totalLines;
    const quint32 avgHsyncHigh = m_sumHsyncHigh/totalLines;

    if (avgHsyncLow > avgHsyncHigh) {
        m_hSyncPolarity = !m_hSyncPolarity;
        qDebug()<<"Swap HSYNC Polarity"<<m_hSyncPolarity;
        return;
    }

    if (m_sumVsyncLow == 2 && m_lastvVsyncLineCounter == 522 && avgHsyncLow == 300 && avgHsyncHigh == 2260) {
        qDebug()<<"VGA 640x480";
        m_nextImage = QImage(640,480,QImage::Format_Indexed8);
        m_vFrontPorch = -31;
        m_hFrontPorch = -54;
        m_hNominator = 5;
        m_hDenominator = 16;
        m_nextImage.fill(0);
        m_nextImage.setColorTable(m_colorTable);
        return;
    }
    if (m_sumVsyncLow == 2 && m_lastvVsyncLineCounter == 523 && avgHsyncLow >= 300 && avgHsyncLow <= 304 && avgHsyncHigh == 2236) {
        qDebug()<<"VGA 640x480 ALT";
        m_nextImage = QImage(640,480,QImage::Format_Indexed8);
        m_vFrontPorch = -31;
        m_hFrontPorch = -54;
        m_hNominator = 5;
        m_hDenominator = 16;
        m_nextImage.fill(0);
        m_nextImage.setColorTable(m_colorTable);
        return;
    }
    if (m_sumVsyncLow == 6 && m_lastvVsyncLineCounter == 800 && avgHsyncLow == 167 && avgHsyncHigh == 1486) {
        qDebug()<<"VGA 1024x768";
        m_nextImage = QImage(1024,768,QImage::Format_Indexed8);
        m_vFrontPorch = -29;
        m_hFrontPorch = -161;
        m_hNominator = 256;
        m_hDenominator = 315;
        m_nextImage.fill(0);
        m_nextImage.setColorTable(m_colorTable);
        return;
    }

    qDebug()<<"Invalid Frame Format VSYNC INFO"<<m_sumVsyncLow<<m_lastvVsyncLineCounter<<avgHsyncLow<<avgHsyncHigh;
    m_nextImage = QImage();
}

void VGADecoder::serialize(QDataStream& stream) const {
    HWTraceable::serialize(stream);
    stream<<m_nextImage;
    stream<<m_colorTable;
    stream<<m_timeCode;
    stream<<m_firstPinNumber;
    stream<<m_vSyncPinMask;
    stream<<m_hSyncPinMask;
    stream<<m_lastHsyncTimeStamp;
    stream<<m_lastvVsyncLineCounter;
    stream<<m_sumHsyncLow;
    stream<<m_sumHsyncHigh;
    stream<<m_sumVsyncLow;
    stream<<m_frameCounter;
    stream<<m_hNominator;
    stream<<m_hDenominator;
    stream<<m_hFrac;
    stream<<m_x;
    stream<<m_y;
    stream<<m_vFrontPorch;
    stream<<m_hFrontPorch;
    stream<<m_lastVsync;
    stream<<m_lastHsync;
    stream<<m_frozenVSync;
    stream<<m_hSyncPolarity;
    stream<<m_vSyncPolarity;
}

void VGADecoder::deserialize(QDataStream& stream) {
    HWTraceable::deserialize(stream);
    stream>>m_nextImage;
    stream>>m_colorTable;
    stream>>m_timeCode;
    stream>>m_firstPinNumber;
    stream>>m_vSyncPinMask;
    stream>>m_hSyncPinMask;
    stream>>m_lastHsyncTimeStamp;
    stream>>m_lastvVsyncLineCounter;
    stream>>m_sumHsyncLow;
    stream>>m_sumHsyncHigh;
    stream>>m_sumVsyncLow;
    stream>>m_frameCounter;
    stream>>m_hNominator;
    stream>>m_hDenominator;
    stream>>m_hFrac;
    stream>>m_x;
    stream>>m_y;
    stream>>m_vFrontPorch;
    stream>>m_hFrontPorch;
    stream>>m_lastVsync;
    stream>>m_lastHsync;
    stream>>m_frozenVSync;
    stream>>m_hSyncPolarity;
    stream>>m_vSyncPolarity;
}
